@@ -13,10 +13,30 @@
 """SSH agent client"""
 
 import asyncio
+import errno
 import os
+import sys
+import tempfile
 
 import asyncssh
 
+from .logging import logger
+
+
+try:
+    if sys.platform == 'win32': # pragma: no cover
+        from .agent_win32 import open_agent
+    else:
+        from .agent_unix import open_agent
+except ImportError as exc: # pragma: no cover
+    def open_agent(loop, agent_path, reason=str(exc)):
+        """Dummy function if we're unable to import agent support"""
+
+        # pylint: disable=unused-argument
+
+        raise OSError(errno.ENOENT, 'Agent support unavailable: %s' % reason)
+
+from .listener import create_unix_forward_listener
 from .misc import ChannelOpenError, load_default_keypairs
 from .packet import Byte, String, UInt32, PacketDecodeError, SSHPacket
 from .public_key import SSHKeyPair
@@ -57,6 +77,26 @@ SSH_AGENT_RSA_SHA2_512                   = 4
 # pylint: enable=bad-whitespace
 
 
+class _X11AgentListener:
+    """Listener used to forward agent connections"""
+
+    def __init__(self, tempdir, path, unix_listener):
+        self._tempdir = tempdir
+        self._path = path
+        self._unix_listener = unix_listener
+
+    def get_path(self):
+        """Return the path being listened on"""
+
+        return self._path
+
+    def close(self):
+        """Close the agent listener"""
+
+        self._unix_listener.close()
+        self._tempdir.cleanup()
+
+
 class SSHAgentKeyPair(SSHKeyPair):
     """Surrogate for a key managed by the SSH agent"""
 
@@ -122,7 +162,7 @@ class SSHAgentClient:
         self._agent_path = agent_path
         self._reader = None
         self._writer = None
-        self._lock = asyncio.Lock()
+        self._lock = asyncio.Lock(loop=loop)
 
     def _cleanup(self):
         """Clean up this SSH agent client"""
@@ -150,15 +190,12 @@ class SSHAgentClient:
     def connect(self):
         """Connect to the SSH agent"""
 
-        if isinstance(self._agent_path, str):
-            # pylint doesn't think open_unix_connection exists
-            # pylint: disable=no-member
-            self._reader, self._writer = \
-                yield from asyncio.open_unix_connection(self._agent_path,
-                                                        loop=self._loop)
-        else:
+        if isinstance(self._agent_path, asyncssh.SSHServerConnection):
             self._reader, self._writer = \
                 yield from self._agent_path.open_agent_connection()
+        else:
+            self._reader, self._writer = \
+                yield from open_agent(self._loop, self._agent_path)
 
     @asyncio.coroutine
     def _make_request(self, msgtype, *args):
@@ -253,19 +290,21 @@ class SSHAgentClient:
                :file:`.ssh/id_ed25519-cert.pub`,
                :file:`.ssh/id_ecdsa-cert.pub`, :file:`.ssh/id_rsa-cert.pub`,
                and :file:`.ssh/id_dsa-cert.pub`.
-           :param str passphrase: (optional)
+           :param passphrase: (optional)
                The passphrase to use to decrypt the keys.
            :param lifetime: (optional)
                The time in seconds after which the keys should be
-               automatically deleted, or ``None`` to store these keys
+               automatically deleted, or `None` to store these keys
                indefinitely (the default).
-           :param bool confirm: (optional)
+           :param confirm: (optional)
                Whether or not to require confirmation for each private
-               key operation which uses these keys, defaulting to ``False``.
+               key operation which uses these keys, defaulting to `False`.
            :type keylist: *see* :ref:`SpecifyingPrivateKeys`
-           :type lifetime: `int` or ``None``
+           :type passphrase: `str`
+           :type lifetime: `int` or `None`
+           :type confirm: `bool`
 
-           :raises: ::exc::`ValueError` if the keys cannot be added
+           :raises: :exc:`ValueError` if the keys cannot be added
 
         """
 
@@ -279,11 +318,11 @@ class SSHAgentClient:
                       SSH_AGENTC_ADD_IDENTITY
 
         for keypair in keypairs:
-            comment = keypair.get_comment()
+            comment = keypair.get_comment_bytes()
             resptype, resp = \
                 yield from self._make_request(msgtype,
                                               keypair.get_agent_private_key(),
-                                              String(comment or ''),
+                                              String(comment or b''),
                                               constraints)
 
             if resptype == SSH_AGENT_SUCCESS:
@@ -298,21 +337,23 @@ class SSHAgentClient:
                            lifetime=None, confirm=False):
         """Store keys associated with a smart card in the agent
 
-           :param str provider:
+           :param provider:
                The name of the smart card provider
            :param pin: (optional)
                The PIN to use to unlock the smart card
            :param lifetime: (optional)
                The time in seconds after which the keys should be
-               automatically deleted, or ``None`` to store these keys
+               automatically deleted, or `None` to store these keys
                indefinitely (the default).
-           :param bool confirm: (optional)
+           :param confirm: (optional)
                Whether or not to require confirmation for each private
-               key operation which uses these keys, defaulting to ``False``.
-           :type pin: `str` or ``None``
-           :type lifetime: `int` or ``None``
+               key operation which uses these keys, defaulting to `False`.
+           :type provider: `str`
+           :type pin: `str` or `None`
+           :type lifetime: `int` or `None`
+           :type confirm: `bool`
 
-           :raises: ::exc::`ValueError` if the keys cannot be added
+           :raises: :exc:`ValueError` if the keys cannot be added
 
         """
 
@@ -337,9 +378,9 @@ class SSHAgentClient:
 
            :param keylist:
                The list of keys to remove.
-           :type keylist: list of :class:`SSHKeyPair`
+           :type keylist: `list` of :class:`SSHKeyPair`
 
-           :raises: ::exc::`ValueError` if any keys are not found
+           :raises: :exc:`ValueError` if any keys are not found
 
         """
 
@@ -359,13 +400,14 @@ class SSHAgentClient:
     def remove_smartcard_keys(self, provider, pin=None):
         """Remove keys associated with a smart card stored in the agent
 
-           :param str provider:
+           :param provider:
                The name of the smart card provider
            :param pin: (optional)
                The PIN to use to unlock the smart card
-           :type pin: `str` or ``None``
+           :type provider: `str`
+           :type pin: `str` or `None`
 
-           :raises: ::exc::`ValueError` if the keys are not found
+           :raises: :exc:`ValueError` if the keys are not found
 
         """
 
@@ -384,7 +426,7 @@ class SSHAgentClient:
     def remove_all(self):
         """Remove all keys stored in the agent
 
-           :raises: ::exc::`ValueError` if the keys can't be removed
+           :raises: :exc:`ValueError` if the keys can't be removed
 
         """
 
@@ -402,10 +444,11 @@ class SSHAgentClient:
     def lock(self, passphrase):
         """Lock the agent using the specified passphrase
 
-           :param str passphrase:
+           :param passphrase:
                The passphrase required to later unlock the agent
+           :type passphrase: `str`
 
-           :raises: ::exc::`ValueError` if the agent can't be locked
+           :raises: :exc:`ValueError` if the agent can't be locked
 
         """
 
@@ -423,10 +466,11 @@ class SSHAgentClient:
     def unlock(self, passphrase):
         """Unlock the agent using the specified passphrase
 
-           :param str passphrase:
+           :param passphrase:
                The passphrase to use to unlock the agent
+           :type passphrase: `str`
 
-           :raises: ::exc::`ValueError` if the agent can't be unlocked
+           :raises: :exc:`ValueError` if the agent can't be unlocked
 
         """
 
@@ -474,7 +518,7 @@ class SSHAgentClient:
         """Close the SSH agent connection
 
            This method closes the connection to the ssh-agent. Any
-           attempts to use this :class:``SSHAgentClient`` or the key
+           attempts to use this :class:`SSHAgentClient` or the key
            pairs it previously returned will result in an error.
 
         """
@@ -487,15 +531,15 @@ def connect_agent(agent_path=None, *, loop=None):
     """Make a connection to the SSH agent
 
        This function attempts to connect to an ssh-agent process
-       listening on a UNIX domain socket at ``agent_path``. If not
-       provided, it will attempt to get the path from the ``SSH_AUTH_SOCK``
+       listening on a UNIX domain socket at `agent_path`. If not
+       provided, it will attempt to get the path from the `SSH_AUTH_SOCK`
        environment variable.
 
-       If the connection is successful, an ``SSHAgentClient`` object
+       If the connection is successful, an :class:`SSHAgentClient` object
        is returned that has methods on it you can use to query the
        ssh-agent. If no path is specified and the environment variable
        is not set or the connection to the agent fails, this function
-       returns ``None``.
+       returns `None`.
 
        :param agent_path: (optional)
            The path to use to contact the ssh-agent process, or the
@@ -504,25 +548,33 @@ def connect_agent(agent_path=None, *, loop=None):
        :param loop: (optional)
            The event loop to use when creating the connection. If not
            specified, the default event loop is used.
-       :type agent_path: str or :class:`SSHServerConnection`
+       :type agent_path: `str` or :class:`SSHServerConnection`
+       :type loop: :class:`AbstractEventLoop <asyncio.AbstractEventLoop>`
 
-       :returns: An :class:`SSHAgentClient` or ``None``
+       :returns: An :class:`SSHAgentClient` or `None`
 
     """
 
-    if not loop:
-        loop = asyncio.get_event_loop()
-
-    if not agent_path:
-        agent_path = os.environ.get('SSH_AUTH_SOCK', None)
-
-        if not agent_path:
-            return None
-
     agent = SSHAgentClient(loop, agent_path)
 
     try:
         yield from agent.connect()
         return agent
-    except (OSError, ChannelOpenError):
+    except (OSError, ChannelOpenError) as exc:
+        logger.warning('Unable to contact agent: %s', exc)
+        return None
+
+
+@asyncio.coroutine
+def create_agent_listener(conn, loop):
+    """Create a listener for forwarding ssh-agent connections"""
+
+    try:
+        tempdir = tempfile.TemporaryDirectory(prefix='asyncssh-')
+        path = os.path.join(tempdir.name, 'agent')
+        unix_listener = yield from create_unix_forward_listener(
+            conn, loop, conn.create_agent_connection, path)
+
+        return _X11AgentListener(tempdir, path, unix_listener)
+    except OSError:
         return None
